# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from google.pubsub_v1 import ReceivedMessage

from aorta.types import Acknowledgable


class Frame(Acknowledgable):

    def __init__(self, frame: ReceivedMessage) -> None:
        self.data = frame.message.data
        self.frame = frame
        self.message_id = frame.message.message_id

    def ack(self):
        pass

    def nack(self):
        pass